import sys
import warnings
import scipy
import scipy.sparse
import numpy as np
import numpy.linalg


def kernel(x, xKern, theta):
    # KERNEL Compute the rbf kernel
    n2 = dist2(x, xKern)
    if (theta[0] == 0):
        theta[0] = 2 / np.median(n2[np.where(np.tril(n2) > 0)])
        theta_new = theta[0]
    wi2 = theta[0] / 2
    kx = theta[1] * np.exp(-n2 * wi2)
    bw_new = 1 / theta[0]
    return kx, bw_new


def dist2(x, c):
    # DIST2	Calculates squared distance between two sets of points.
    #
    # Description
    # D = DIST2(X, C) takes two matrices of vectors and calculates the
    # squared Euclidean distance between them.  Both matrices must be of
    # the same column dimension.  If X has M rows and N columns, and C has
    # L rows and N columns, then the result has M rows and L columns.  The
    # I, Jth entry is the  squared distance from the Ith row of X to the
    # Jth row of C.
    #
    # See also
    # GMMACTIV, KMEANS, RBFFWD
    #

    ndata, dimx = x.shape
    ncentres, dimc = c.shape
    if (dimx != dimc):
        raise Exception('Data dimension does not match dimension of centres')

    n2 = (np.matlib.ones((ncentres, 1)) * np.sum(np.multiply(x, x).T, axis=0)).T + \
         np.matlib.ones((ndata, 1)) * np.sum(np.multiply(c, c).T, axis=0) - \
         2 * (x * c.T)

    # Rounding errors occasionally cause negative entries in n2
    n2[np.where(n2 < 0)] = 0
    return n2


def pdinv(A):
    # PDINV Computes the inverse of a positive definite matrix
    numData = A.shape[0]
    try:
        U = np.linalg.cholesky(A).T
        invU = np.eye(numData).dot(np.linalg.inv(U))
        Ainv = invU.dot(invU.T)
    except numpy.linalg.LinAlgError as e:
        warnings.warn('Matrix is not positive definite in pdinv, inverting using svd')
        u, s, vh = np.linalg.svd(A, full_matrices=True)
        Ainv = vh.T.dot(np.diag(1 / s)).dot(u.T)
    except Exception as e:
        raise e
    return np.matlib.asmatrix(Ainv)


def eigdec(x, N, evals_only=False):
    # EIGDEC	Sorted eigendecomposition
    #
    #	Description
    #	 EVALS = EIGDEC(X, N computes the largest N eigenvalues of the
    #	matrix X in descending order.  [EVALS, EVEC] = EIGDEC(X, N) also
    #	computes the corresponding eigenvectors.
    #
    #	See also
    #	PCA, PPCA
    #

    if (N != np.round(N) or N < 1 or N > x.shape[1]):
        raise Exception('Number of PCs must be integer, >0, < dim')

    # Find the eigenvalues of the data covariance matrix
    if (evals_only):
        # Use eig function as always more efficient than eigs here
        temp_evals, _ = np.linalg.eig(x)
    else:
        # Use eig function unless fraction of eigenvalues required is tiny
        if ((N / x.shape[1]) > 0.04):
            temp_evals, temp_evec = np.linalg.eig(x)
        else:
            temp_evals, temp_evec = scipy.sparse.linalg.eigs(x, k=N, which='LM')

    # Eigenvalues nearly always returned in descending order, but just to make sure.....
    evals = np.sort(-temp_evals)
    perm = np.argsort(-temp_evals)
    evals = -evals[0:N]

    if (not evals_only):
        if (np.all(evals == temp_evals[0:N])):
            # Originals were in order
            evec = temp_evec[:, 0: N]
        else:
            # Need to reorder the eigenvectors
            evec = np.empty_like(temp_evec[:, 0: N])
            for i in range(N):
                evec[:, i] = temp_evec[:, perm[i]]

        return evals.astype(float), evec.astype(float)
    else:
        return evals.astype(float)


def minimize(X, f, length, *varargin):
    # Minimize a differentiable multivariate function.
    #
    # Usage: X, fX, i = minimize(X, f, length, P1, P2, P3, ... )
    #
    # where the starting point is given by "X" (D by 1), and the function named in
    # the string "f", must return a function value and a vector of partial
    # derivatives of f wrt X, the "length" gives the length of the run: if it is
    # positive, it gives the maximum number of line searches, if negative its
    # absolute gives the maximum allowed number of function evaluations. You can
    # (optionally) give "length" a second component, which will indicate the
    # reduction in function value to be expected in the first line-search (defaults
    # to 1.0). The parameters P1, P2, P3, ... are passed on to the function f.
    #
    # The function returns when either its length is up, or if no further progress
    # can be made (ie, we are at a (local) minimum, or so close that due to
    # numerical problems, we cannot get any closer). NOTE: If the function
    # terminates within a few iterations, it could be an indication that the
    # function values and derivatives are not consistent (ie, there may be a bug in
    # the implementation of your "f" function). The function returns the found
    # solution "X", a vector of function values "fX" indicating the progress made
    # and "i" the number of iterations (line searches or function evaluations,
    # depending on the sign of "length") used.
    #
    # The Polack-Ribiere flavour of conjugate gradients is used to compute search
    # directions, and a line search using quadratic and cubic polynomial
    # approximations and the Wolfe-Powell stopping criteria is used together with
    # the slope ratio method for guessing initial step sizes. Additionally a bunch
    # of checks are made to make sure that exploration is taking place and that
    # extrapolation will not be unboundedly large.
    #
    # See also: checkgrad

    INT = 0.1  # don't reevaluate within 0.1 of the limit of the current bracket
    EXT = 3.0  # extrapolate maximum 3 times the current step-size
    MAX = 20  # max 20 function evaluations per line search
    RATIO = 10  # maximum allowed slope ratio
    SIG = 0.1
    RHO = SIG / 2  # SIG and RHO are the constants controlling the Wolfe-
    # Powell conditions. SIG is the maximum allowed absolute ratio between
    # previous and new slopes (derivatives in the search direction), thus setting
    # SIG to low (positive) values forces higher precision in the line-searches.
    # RHO is the minimum allowed fraction of the expected (from the slope at the
    # initial point in the linesearch). Constants must satisfy 0 < RHO < SIG < 1.
    # Tuning of SIG (depending on the nature of the function to be optimized) may
    # speed up the minimization; it is probably not worth playing much with RHO.

    # The code falls naturally into 3 parts, after the initial line search is
    # started in the direction of steepest descent. 1) we first enter a while loop
    # which uses point 1 (p1) and (p2) to compute an extrapolation (p3), until we
    # have extrapolated far enough (Wolfe-Powell conditions). 2) if necessary, we
    # enter the second loop which takes p2, p3 and p4 chooses the subinterval
    # containing a (local) minimum, and interpolates it, unil an acceptable point
    # is found (Wolfe-Powell conditions). Note, that points are always maintained
    # in order p0 <= p1 <= p2 < p3 < p4. 3) compute a new search direction using
    # conjugate gradients (Polack-Ribiere flavour), or revert to steepest if there
    # was a problem in the previous line-search. Return the best value so far, if
    # two consecutive line-searches fail, or whenever we run out of function
    # evaluations or line-searches. During extrapolation, the "f" function may fail
    # either with an error or returning Nan or Inf, and minimize should handle this
    # gracefully.

    if np.size(length) == 2:
        red = length[1]
        length = length[0]
    else:
        red = 1

    if length > 0:
        S = 'Linesearch'
    else:
        S = 'Function evaluation'

    i = 0  # zero the run length counter
    ls_failed = 0  # no previous line search has failed
    temp = [f, X]
    temp.extend(varargin)
    temp.extend([None, 2])
    f0, df0 = feval(temp)  # get function value and gradient
    fX = f0
    i = i + (1 if length < 0 else 0)  # count epochs?!
    s = -df0
    d0 = (-s.T * s)[0, 0]  # initial search direction (steepest) and slope
    x3 = red / (1 - d0)  # initial step is red/(|s|+1)

    while i < abs(length):  # while not finished
        i = i + (1 if length > 0 else 0)  # count iterations?!
        X0 = X  # make a copy of current values
        F0 = f0
        dF0 = df0
        if length > 0:
            M = MAX
        else:
            M = min(MAX, -length - i)

        while 1:  # keep extrapolating as long as necessary
            x2 = 0
            f2 = f0
            d2 = d0
            f3 = f0
            df3 = df0
            success = False

            while (not success and M > 0):
                try:
                    M = M - 1
                    i = i + (1 if length < 0 else 0)  # count epochs?!
                    temp = [f, X + x3 * s]
                    temp.extend(varargin)
                    temp.extend([None, 2])
                    f3, df3 = feval(temp)
                    if np.isnan(f3) or np.isinf(f3) or np.any(np.isnan(df3)) or np.any(np.isinf(df3)):
                        raise Exception('')
                    success = True
                except Exception as e:  # catch any error which occured in f
                    x3 = (x2 + x3) / 2  # bisect and try again

            if f3 < F0:
                X0 = X + x3 * s  # keep best values
                F0 = f3
                dF0 = df3
            d3 = (df3.T * s)[0, 0]  # new slope
            if d3 > SIG * d0 or f3 > f0 + x3 * RHO * d0 or M == 0:  # are we done extrapolating?
                break

            x1 = x2  # move point 2 to point 1
            f1 = f2
            d1 = d2
            x2 = x3  # move point 3 to point 2
            f2 = f3
            d2 = d3
            A = 6 * (f1 - f2) + 3 * (d2 + d1) * (x2 - x1)  # make cubic extrapolation
            B = 3 * (f2 - f1) - (2 * d1 + d2) * (x2 - x1)
            if B * B - A * d1 * (x2 - x1) < 0:
                x3 = x1 - d1 * (x2 - x1) ** 2 / B # TODO: Is it reasonable?
            else:
                x3 = x1 - d1 * (x2 - x1) ** 2 / (B + np.sqrt(B * B - A * d1 * (x2 - x1)))  # num. error possible, ok!

            if not np.isreal(x3) or np.isnan(x3) or np.isinf(x3) or x3 < 0:  # num prob | wrong sign?
                x3 = x2 * EXT  # extrapolate maximum amount
            elif x3 > x2 * EXT:  # new point beyond extrapolation limit?
                x3 = x2 * EXT  # extrapolate maximum amount
            elif x3 < x2 + INT * (x2 - x1):  # new point too close to previous point?
                x3 = x2 + INT * (x2 - x1)
        # end extrapolation

        while (abs(d3) > -SIG * d0 or f3 > f0 + x3 * RHO * d0) and M > 0:  # keep interpolating
            if d3 > 0 or f3 > f0 + x3 * RHO * d0:  # choose subinterval
                x4 = x3  # move point 3 to point 4
                f4 = f3
                d4 = d3
            else:
                x2 = x3  # move point 3 to point 2
                f2 = f3
                d2 = d3

            if f4 > f0:
                x3 = x2 - (0.5 * d2 * (x4 - x2) ** 2) / (f4 - f2 - d2 * (x4 - x2))  # quadratic interpolation
            else:
                A = 6 * (f2 - f4) / (x4 - x2) + 3 * (d4 + d2)  # cubic interpolation
                B = 3 * (f4 - f2) - (2 * d2 + d4) * (x4 - x2)
                x3 = x2 + (np.sqrt(B * B - A * d2 * (x4 - x2) ** 2) - B) / A  # num. error possible, ok!

            if np.isnan(x3) or np.isinf(x3):
                x3 = (x2 + x4) / 2  # if we had a numerical problem then bisect

            x3 = max(min(x3, x4 - INT * (x4 - x2)), x2 + INT * (x4 - x2))  # don't accept too close
            temp = [f, X + x3 * s]
            temp.extend(varargin)
            temp.extend([None, 2])
            f3, df3 = feval(temp)
            if f3 < F0:
                X0 = X + x3 * s  # keep best values
                F0 = f3
                dF0 = df3
            M = M - 1
            i = i + (1 if length < 0 else 0)  # count epochs?!
            d3 = (df3.T * s)[0, 0]  # new slope
        # end interpolation

        if (abs(d3) < -SIG * d0 and f3 < f0 + x3 * RHO * d0):  # if line search succeeded
            X = X + x3 * s
            f0 = f3
            fX = np.vstack([fX, f0])  # update variables
            s = ((df3.T * df3)[0, 0] - df0.T * df3[0, 0]) / (df0.T * df0)[0, 0] * s - df3  # Polack-Ribiere CG direction
            df0 = df3  # swap derivatives
            d3 = d0
            d0 = (df0.T * s)[0, 0]
            if (d0 > 0):  # new slope must be negative
                s = -df0
                d0 = -(s.T * s)[0, 0]
            x3 = x3 * min(RATIO, d3 / (d0 - sys.float_info.min))  # slope ratio but max RATIO
            ls_failed = 0  # this line search did not fail
        else:
            X = X0  # restore best point so far
            f0 = F0
            df0 = dF0
            if (ls_failed or i > abs(length)):  # line search failed twice in a row
                break  # or we ran out of time, so we give up
            s = -df0  # try steepest
            d0 = -(s.T * s)[0, 0]
            x3 = 1 / (1 - d0)
            ls_failed = 1  # this line search failed
    return X, fX, i


def feval(parameters):
    if parameters[0] == 'covSum':
        if (len(parameters) == 1):
            return cov_sum()
        elif (len(parameters) == 2):
            return cov_sum(parameters[1])
        elif (len(parameters) == 3):
            return cov_sum(parameters[1], parameters[2])
        elif (len(parameters) == 4):
            return cov_sum(parameters[1], parameters[2], parameters[3])
        elif (len(parameters) == 5):
            return cov_sum(parameters[1], parameters[2], parameters[3], parameters[4])
        elif (len(parameters) == 6):
            return cov_sum(parameters[1], parameters[2], parameters[3], parameters[4], parameters[5])
    elif parameters[0] == 'covNoise':
        if (len(parameters) == 1):
            return cov_noise()
        elif (len(parameters) == 2):
            return cov_noise(parameters[1])
        elif (len(parameters) == 3):
            return cov_noise(parameters[1], parameters[2])
        elif (len(parameters) == 4):
            return cov_noise(parameters[1], parameters[2], parameters[3])
        else:
            return cov_noise(parameters[1], parameters[2], parameters[3], parameters[4])
    elif parameters[0] == 'covSEard':
        if (len(parameters) == 1):
            return cov_seard()
        elif (len(parameters) == 2):
            return cov_seard(parameters[1])
        elif (len(parameters) == 3):
            return cov_seard(parameters[1], parameters[2])
        elif (len(parameters) == 4):
            return cov_seard(parameters[1], parameters[2], parameters[3])
        else:
            return cov_seard(parameters[1], parameters[2], parameters[3], parameters[4])
    elif parameters[0] == 'covSum':
        if (len(parameters) == 1):
            return cov_sum()
        elif (len(parameters) == 2):
            return cov_sum(parameters[1])
        elif (len(parameters) == 3):
            return cov_sum(parameters[1], parameters[2])
        elif (len(parameters) == 4):
            return cov_sum(parameters[1], parameters[2], parameters[3])
        elif (len(parameters) == 5):
            return cov_sum(parameters[1], parameters[2], parameters[3], parameters[4])
        elif (len(parameters) == 6):
            return cov_sum(parameters[1], parameters[2], parameters[3], parameters[4], parameters[5])
    elif parameters[0] == 'gpr_multi_new':
        if (len(parameters) == 1):
            return gpr_multi_new()
        elif (len(parameters) == 2):
            return gpr_multi_new(parameters[1])
        elif (len(parameters) == 3):
            return gpr_multi_new(parameters[1], parameters[2])
        elif (len(parameters) == 4):
            return gpr_multi_new(parameters[1], parameters[2], parameters[3])
        elif (len(parameters) == 5):
            return gpr_multi_new(parameters[1], parameters[2], parameters[3], parameters[4])
        elif (len(parameters) == 6):
            return gpr_multi_new(parameters[1], parameters[2], parameters[3], parameters[4], parameters[5])
        elif (len(parameters) == 7):
            return gpr_multi_new(parameters[1], parameters[2], parameters[3], parameters[4], parameters[5],
                                 parameters[6])
    else:
        raise Exception('Undefined function')


def gpr_multi_new(logtheta=None, covfunc=None, x=None, y=None, xstar=None, nargout=1):
    # Here we change the function gpr to gpr_multi, in which y contains a set
    # of vectors on which we do repression from x

    # gpr - Gaussian process regression, with a named covariance function. Two
    # modes are possible: training and prediction: if no test data are given, the
    # function returns minus the log likelihood and its partial derivatives with
    # respect to the hyperparameters; this mode is used to fit the hyperparameters.
    # If test data are given, then (marginal) Gaussian predictions are computed,
    # whose mean and variance are returned. Note that in cases where the covariance
    # function has noise contributions, the variance returned in S2 is for noisy
    # test targets; if you want the variance of the noise-free latent function, you
    # must substract the noise variance.
    #
    # usage: [nlml dnlml] = gpr(logtheta, covfunc, x, y)
    #    or: [mu S2]  = gpr(logtheta, covfunc, x, y, xstar)
    #
    # where:
    #
    #   logtheta is a (column) vector of log hyperparameters
    #   covfunc  is the covariance function
    #   x        is a n by D matrix of training inputs
    #   y        is a (column) vector (of size n) of targets
    #   xstar    is a nn by D matrix of test inputs
    #   nlml     is the returned value of the negative log marginal likelihood
    #   dnlml    is a (column) vector of partial derivatives of the negative
    #                 log marginal likelihood wrt each log hyperparameter
    #   mu       is a (column) vector (of size nn) of prediced means
    #   S2       is a (column) vector (of size nn) of predicted variances
    #
    # For more help on covariance functions, see "covFunctions".

    if type(covfunc) == str:
        covfunc = [covfunc]  # convert to cell if needed
    n, D = x.shape
    n, m = y.shape
    if eval(feval(covfunc)) != logtheta.shape[0]:
        raise Exception('Error: Number of parameters do not agree with covariance function')

    temp = list(covfunc.copy())
    temp.append(logtheta)
    temp.append(x)
    K = feval(temp)  # compute training set covariance matrix

    L = np.linalg.cholesky(K)  # cholesky factorization of the covariance
    alpha = solve_chol(L.T, y)

    if (
            logtheta is not None and covfunc is not None and x is not None and y is not None and xstar is None):  # if no test cases, compute the negative log marginal likelihood
        out1 = 0.5 * np.trace(y.T * alpha) + m * np.sum(np.log(np.diag(L)), axis=0) + 0.5 * m * n * np.log(
            2 * np.pi)
        if nargout == 2:  # ... and if requested, its partial derivatives
            out2 = np.matlib.zeros((logtheta.shape[0], 1))  # set the size of the derivative vector
            W = m * (np.linalg.inv(L.T) * (
                    np.linalg.inv(L) * np.matlib.eye(n))) - alpha * alpha.T  # precompute for convenience
            for i in range(len(out2) - 1, len(out2)):
                temp = list(covfunc.copy())
                temp.append(logtheta)
                temp.append(x)
                temp.append(i)
                out2[i] = np.sum(np.multiply(W, feval(temp))) / 2
    else:  # ... otherwise compute (marginal) test predictions ...
        temp = list(covfunc.copy())
        temp.append(logtheta)
        temp.append(x)
        temp.append(xstar)
        temp.append(2)  # nargout == 2
        Kss, Kstar = feval(temp)  # test covariances
        out1 = Kstar.T * alpha  # predicted means

        if nargout == 2:
            v = np.linalg.inv(L) * Kstar
            v = np.asmatrix(v)
            out2 = Kss - np.sum(np.multiply(v, v), axis=0).T

    if nargout == 1:
        return out1
    else:
        return out1, out2


def solve_chol(A, B):
    # solve_chol - solve linear equations from the Cholesky factorization.
    # Solve A*X = B for X, where A is square, symmetric, positive definite. The
    # input to the function is R the Cholesky decomposition of A and the matrix B.
    # Example: X = solve_chol(chol(A),B);
    #
    # NOTE: The program code is written in the C language for efficiency and is
    # contained in the file solve_chol.c, and should be compiled using matlabs mex
    # facility. However, this file also contains a (less efficient) matlab
    # implementation, supplied only as a help to people unfamiliar with mex. If
    # the C code has been properly compiled and is avaiable, it automatically
    # takes precendence over the matlab code in this file.

    if A is None or B is None:
        raise Exception('Wrong number of arguments.')

    if (A.shape[0] != A.shape[1] or A.shape[0] != B.shape[0]):
        raise Exception('Wrong sizes of matrix arguments.')

    res = np.linalg.inv(A) * (np.linalg.inv(A.T) * B)
    return res


K = np.matlib.empty((0, 0))


def cov_noise(logtheta=None, x=None, z=None, nargout=1):
    # Independent covariance function, ie "white noise", with specified variance.
    # The covariance function is specified as:
    #
    # k(x^p,x^q) = s2 * \delta(p,q)
    #
    # where s2 is the noise variance and \delta(p,q) is a Kronecker delta function
    # which is 1 iff p=q and zero otherwise. The hyperparameter is
    #
    # logtheta = [ log(sqrt(s2)) ]
    #
    # For more help on design of covariance functions, see "covFunctions".

    if (logtheta is None and x is None and z is None):  # report number of parameters
        A = '1'

        return A

    s2 = np.exp(2 * logtheta)[0, 0]  # noise variance

    if (logtheta is not None and x is not None and z is None):  # compute covariance matrix
        A = s2 * np.matlib.eye(x.shape[0])
    elif (nargout == 2):  # compute test set covariances
        A = s2
        B = 0  # zeros cross covariance by independence
    else:  # compute derivative matrix
        A = 2 * s2 * np.matlib.eye(x.shape[0])

    if (nargout == 2):
        return A, B
    else:
        return A


def cov_seard(loghyper=None, x=None, z=None, nargout=1):
    # Squared Exponential covariance function with Automatic Relevance Detemination
    # (ARD) distance measure. The covariance function is parameterized as:
    #
    # k(x^p,x^q) = sf2 * exp(-(x^p - x^q)'*inv(P)*(x^p - x^q)/2)
    #
    # where the P matrix is diagonal with ARD parameters ell_1^2,...,ell_D^2, where
    # D is the dimension of the input space and sf2 is the signal variance. The
    # hyperparameters are:
    #
    # loghyper = [ log(ell_1)
    #              log(ell_2)
    #               .
    #              log(ell_D)
    #              log(sqrt(sf2)) ]
    #
    # For more help on design of covariance functions, see "covFunctions".
    global K

    if (loghyper is None and x is None and z is None):
        A = '(D+1)'

        return A  # report number of parameters

    n, D = x.shape
    loghyper = loghyper.T.tolist()[0]
    ell = np.exp(loghyper[0:D])  # characteristic length scale
    sf2 = np.exp(2 * loghyper[D])  # signal variance

    if (loghyper is not None and x is not None):
        K = sf2 * np.exp(-sq_dist(np.matlib.diag(1 / ell) * x.T) / 2)
        A = K
    elif nargout == 2:  # compute test set covariances
        A = sf2 * np.matlib.ones((z, 1))
        B = sf2 * np.exp(-sq_dist(np.matlib.diag(1 / ell) * x.T, np.matlib.diag(1 / ell) * z) / 2)
    else:
        # check for correct dimension of the previously calculated kernel matrix
        if (K.shape[0] != n or K.shape[1] != n):
            K = sf2 * np.exp(-sq_dist(np.matlib.diag(1 / ell) * x.T) / 2)

        if z <= D:  # length scale parameters
            A = np.multiply(K, sq_dist(x[:, z].T / ell[z]))
        else:  # magnitude parameter
            A = 2 * K
            K = np.matlib.empty((0, 0))

    if (nargout == 2):
        return A, B
    else:
        return A


def sq_dist(a, b=None, Q=None):
    # sq_dist - a function to compute a matrix of all pairwise squared distances
    # between two sets of vectors, stored in the columns of the two matrices, a
    # (of size D by n) and b (of size D by m). If only a single argument is given
    # or the second matrix is empty, the missing matrix is taken to be identical
    # to the first.
    #
    # Special functionality: If an optional third matrix argument Q is given, it
    # must be of size n by m, and in this case a vector of the traces of the
    # product of Q' and the coordinatewise squared distances is returned.
    #
    # NOTE: The program code is written in the C language for efficiency and is
    # contained in the file sq_dist.c, and should be compiled using matlabs mex
    # facility. However, this file also contains a (less efficient) matlab
    # implementation, supplied only as a help to people unfamiliar with mex. If
    # the C code has been properly compiled and is avaiable, it automatically
    # takes precendence over the matlab code in this file.
    #
    # Usage: C = sq_dist(a, b)
    #    or: C = sq_dist(a)  or equiv.: C = sq_dist(a, [])
    #    or: c = sq_dist(a, b, Q)
    # where the b matrix may be empty.
    #
    # where a is of size D by n, b is of size D by m (or empty), C and Q are of
    # size n by m and c is of size D by 1.

    if b is None or len(b) == 0:  # input arguments are taken to be identical if b is missing or empty
        b = a

    D, n = a.shape
    d, m = b.shape

    if d != D:
        raise Exception('Error: column lengths must agree.')

    if Q is None:
        C = np.matlib.zeros((n, m))
        for d in range(D):
            temp = np.tile(b[d, :], (n, 1)) - np.tile(a[d, :].T, (1, m))
            C = C + np.multiply(temp, temp)
    else:
        if (n, m) == Q.shape:
            C = np.matlib.zeros((D, 1))
            for d in range(D):
                temp = np.tile(b[d, :], (n, 1)) - np.tile(a[d, :].T, (1, m))
                temp = np.multiply(temp, temp)
                temp = np.multiply(temp, Q)
                C[d] = np.sum(temp)
        else:
            raise Exception('Third argument has wrong size.')
    return C


def cov_sum(covfunc, logtheta=None, x=None, z=None, nargout=1):
    # covSum - compose a covariance function as the sum of other covariance
    # functions. This function doesn't actually compute very much on its own, it
    # merely does some bookkeeping, and calls other covariance functions to do the
    # actual work.
    #
    # For more help on design of covariance functions, see "covFunctions".

    j = []
    for i in range(len(covfunc)):  # iterate over covariance functions
        f = covfunc[i]
        j.append([feval([f])])

    if (logtheta is None and x is None and z is None):  # report number of parameters
        A = j[0][0]
        for i in range(1, len(covfunc)):
            A = A + '+' + j[i][0]

        return A

    n, D = x.shape

    v = []  # v vector indicates to which covariance parameters belong
    for i in range(len(covfunc)):
        for k in range(eval(j[i][0])):
            v.append(i)
    v = np.asarray(v)

    if (logtheta is not None and x is not None and z is None):  # compute covariance matrix
        A = np.matlib.zeros((n, n))  # allocate space for covariance matrix
        for i in range(len(covfunc)):  # iteration over summand functions
            f = covfunc[i]
            temp = [f]
            t = logtheta[np.where(v == i)]
            temp.append(t[0] if len(t) == 1 else t)
            temp.append(x)
            A = A + feval(temp)

    if (
            logtheta is not None and x is not None and z is not None):  # compute derivative matrix or test set covariances
        if nargout == 2:  # compute test set cavariances
            A = np.matlib.zeros((z, 1))
            B = np.matlib.zeros((x.shape[0], z))  # allocate space
            for i in range(len(covfunc)):
                f = covfunc[i]
                temp = [f]
                t = logtheta[np.where(v == i)]
                temp.append(t[0] if len(t) == 1 else t)
                temp.append(x)
                temp.append(z)
                temp.append(2)
                AA, BB = feval(temp)  # compute test covariances and accumulate
                A = A + AA
                B = B + BB
        else:  # compute derivative matrices
            i = v[z]  # which covariance function
            j = np.sum(np.where(v[0:z] == i, 1, 0))  # which parameter in that covariance
            f = covfunc[i]
            temp = [f]
            t = logtheta[np.where(v == i)]
            temp.append(t[0] if len(t) == 1 else t)
            temp.append(x)
            temp.append(j)
            A = feval(temp)

    if (nargout == 2):
        return A, B
    else:
        return A



# -*- coding: utf-8 -*-

import hsic
import numpy as np
import itertools
from pygam import LinearGAM
import copy
import pickle

from scipy.stats import gamma
from scipy.stats import scoreatpercentile as sap


# Usage
# 
# execute(X, alpha, num_explanatory_vals) is the method to infer causal graphs.
# 
# X: matrix
# alpha: the alpha level for independence testing
# num_explanatory_vals: the maximum number of variables to infer causal relationships. This is equivalent to d in the paper.
# 
# Return
# P: P[i] contains the indices of the parents of Xi
# U: The indices of variable pairs having UCPs or UBPs
# 
# Before using this, please install numpy and pygam.


def do_an_experiment(data_file, alpha):
	with open(data_file, 'rb') as f:
		obj = pickle.load(f)
	X = obj["data"]
	maxnum_vals = 3
	P, C = execute(X, alpha, maxnum_vals)
	return create_data_for_evaluation(P, C)

def get_neighborhoods(X, alpha):
	n = X.shape[0]
	d = X.shape[1]
	N = [set() for i in range(d)]
	for i in range(d):
			for j in range(d)[i+1:]:
				independence = HSIC.hsic_gam(X=np.reshape(X[:,i], [n, 1]), Y=np.reshape(X[:,j], [n, 1]), mode="pvalue")
				if independence < alpha:
					N[i].add(j)
					N[j].add(i)
	return N



def find_parents(X, alpha, maxnum_vals, N):
	n = X.shape[0]
	d = X.shape[1]
	P = [set() for i in range(d)] # Parents
	t = 2
	Y = copy.deepcopy(X)

	while(True):
		changed = False
		variables_set_list = list(itertools.combinations(set(range(d)),t))
		for variables_set in variables_set_list:
			variables_set = set(variables_set)

			if not check_identified_causality(variables_set, P):
				continue

			child, independence_with_K = get_child(X, variables_set, P, N, Y, alpha)
			if not independence_with_K > alpha:
				continue

			parents = variables_set-{child}
			if not check_independence_withou_K(parents, child, P, N, Y, alpha):
				continue

			for parent in parents:
				P[child].add(parent)
				changed = True
				Y = get_residuals_matrix(X, Y, P, child)

		if changed:
			t = 2
		else:
			t+= 1
			if t > maxnum_vals:
				break

	for i in range(d):
		non_parents = set()
		for j in P[i]:
			residual_i = get_residual(X, i, P[i]-{j})
			residual_j = get_residual(X, j, P[j])
			independence = HSIC.hsic_gam(X=np.reshape(residual_i, [n, 1]), Y=np.reshape(residual_j, [n, 1]), mode="pvalue")
			if independence > alpha:
				non_parents.add(j)
		P[i] = P[i] - non_parents

	return P


def get_residuals_matrix(X, Y_old, P, child):
	n = X.shape[0]
	d = X.shape[1]

	Y = copy.deepcopy(Y_old)
	Y[:,child] = get_residual(X, child, P[child])
	return Y


def get_child(X, variables_set, P, N, Y, alpha):
	n = X.shape[0]
	d = X.shape[1]

	max_independence = 0.0
	max_independence_child = None

	for child in variables_set:
		parents = variables_set - {child}

		if not check_correlation(child, parents, N):
			continue

		residual = get_residual(X, child, parents|P[child])
		independence = HSIC.hsic_gam(X=np.reshape(residual, [n, 1]), Y=np.reshape(Y[:,list(parents)], [n, len(parents)]), mode="pvalue")
		if max_independence < independence:
			max_independence = independence
			max_independence_child = child

	return max_independence_child, max_independence


def check_independence_withou_K(parents, child, P, N, Y, alpha):
	n = Y.shape[0]
	for parent in parents:
		independence = HSIC.hsic_gam(X=np.reshape(Y[:,child], [n, 1]), Y=np.reshape(Y[:,parent], [n, 1]), mode="pvalue")
		if alpha < independence:
			return False
	return True



def check_identified_causality(variables_set, P):
	variables_list = list(variables_set)
	for i in variables_list:
		for j in variables_list[variables_list.index(i)+1:]:
			if (j in P[i]) or (i in P[j]):
				return False
	return True


def check_correlation(child, parents, N):
	for parent in parents:
		if not parent in N[child]:
			return False
	return True


def execute(X, alpha, num_explanatory_vals):

	n = X.shape[0]
	d = X.shape[1]
	N = get_neighborhoods(X, alpha)
	P = find_parents(X, alpha, num_explanatory_vals, N)

	U = []

	for i in range(d):
		for j in range(d)[i+1:]:
			if (i in P[j]) or (j in P[i]):
				continue
			if (not i in N[j]) or (not j in N[i]):
				continue
			i_residual = get_residual(X, i, P[i])
			j_residual = get_residual(X, j, P[j])
			independence = HSIC.hsic_gam(X=np.reshape(i_residual, [n, 1]), Y=np.reshape(j_residual, [n, 1]), mode="pvalue")
			if independence < alpha:
				if not set([i,j]) in U:
					U.append(set([i,j]))

	return P, U


def get_residual(X, explained_i, explanatory_ids):
	n = X.shape[0]

	explanatory_ids = list(explanatory_ids)

	if len(explanatory_ids)==0:
		residual=X[:,explained_i]

	else:
		gam = LinearGAM().fit(X[:,explanatory_ids], X[:,explained_i])

		residual = X[:,explained_i] - gam.predict(X[:,explanatory_ids])

	
	return residual


def create_data_for_evaluation(P, C):
	causal_pairs = []

	for i1 in range(len(P)):
		for i2 in P[i1]:
			causal_pairs.append([i2, i1])

	return causal_pairs, C


##################################################################################
# NOTE: Functions below are for CAMUV.py

def get_width(X):
	n = X.shape[0]

	Xmed = X

	G = np.sum(Xmed * Xmed, 1).reshape(n, 1)
	Q = np.tile(G, (1, n))
	R = np.tile(G.T, (n, 1))

	dists = Q + R - 2 * np.dot(Xmed, Xmed.T)
	dists = dists - np.tril(dists)
	dists = dists.reshape(n ** 2, 1)

	width_x = np.sqrt(0.5 * np.median(dists[dists > 0]))

	return width_x


# def median(lst):
# 	lst_sorted = sorted(lst)
# 	return lst_sorted[(len(lst) - 1) // 2]

def get_mean_width(X):
	n = X.shape[0]

	Xmed = X

	G = np.sum(Xmed * Xmed, 1).reshape(n, 1)
	Q = np.tile(G, (1, n))
	R = np.tile(G.T, (n, 1))

	dists = Q + R - 2 * np.dot(Xmed, Xmed.T)
	dists = dists - np.tril(dists)
	dists = dists.reshape(n ** 2, 1)

	width_x = np.sqrt(0.5 * np.mean(dists[dists > 0]))

	return width_x


def bw_scott(x):
	A = select_sigma(x)
	n = len(x)
	return 1.059 * A * n ** (-0.2)


def bw_silverman(x):
	A = select_sigma(x)
	n = len(x)
	return .9 * A * n ** (-0.2)


def select_sigma(X):
	normalize = 1.349
	IQR = (sap(X, 75) - sap(X, 25)) / normalize
	return np.minimum(np.std(X, axis=0, ddof=1), IQR)


def rbf_dot(pattern1, pattern2, width):
	size1 = pattern1.shape
	size2 = pattern2.shape

	G = np.sum(pattern1 * pattern1, 1).reshape(size1[0], 1)
	H = np.sum(pattern2 * pattern2, 1).reshape(size2[0], 1)

	Q = np.tile(G, (1, size2[0]))
	R = np.tile(H.T, (size1[0], 1))

	H = Q + R - 2 * np.dot(pattern1, pattern2.T)

	H = np.exp(-H / 2 / (width ** 2))

	return H


def get_K(X, width_x):
	n = X.shape[0]

	bone = np.ones((n, 1), dtype=float)
	H = np.identity(n) - np.ones((n, n), dtype=float) / n

	K = rbf_dot(X, X, width_x)
	Kc = np.dot(np.dot(H, K), H)

	return K, Kc


def hsic_gam(X=None, Y=None, alph=None, width_x=None, width_y=None, K=None, Kc=None, L=None, Lc=None, mode=None,
			 kwdth="mdbs"):
	n = X.shape[0]

	if kwdth == "scott":
		width_x = bw_scott(X)
		width_y = bw_scott(Y)
	elif kwdth == "silverman":
		width_x = bw_silverman(X)
		width_y = bw_silverman(Y)
	elif isinstance(kwdth, float):
		width_x = kwdth
		width_y = kwdth
	else:
		if (width_x is None) and ((K is None) or (Kc is None)):
			width_x = get_width(X)
		if (width_y is None) and ((L is None) or (Lc is None)):
			width_y = get_width(Y)

	bone = np.ones((n, 1), dtype=float)
	H = np.identity(n) - np.ones((n, n), dtype=float) / n

	if (K is None) or (Kc is None):
		K = rbf_dot(X, X, width_x)
		Kc = np.dot(np.dot(H, K), H)

	if (L is None) or (Lc is None):
		L = rbf_dot(Y, Y, width_y)
		Lc = np.dot(np.dot(H, L), H)

	testStat = np.sum(Kc.T * Lc) / n

	if mode == "testStat":
		return testStat

	varHSIC = (Kc * Lc / 6) ** 2

	varHSIC = (np.sum(varHSIC) - np.trace(varHSIC)) / n / (n - 1)

	varHSIC = varHSIC * 72 * (n - 4) * (n - 5) / n / (n - 1) / (n - 2) / (n - 3)

	K = K - np.diag(np.diag(K))
	L = L - np.diag(np.diag(L))

	muX = np.dot(np.dot(bone.T, K), bone) / n / (n - 1)
	muY = np.dot(np.dot(bone.T, L), bone) / n / (n - 1)

	mHSIC = (1 + muX * muY - muX - muY) / n

	al = mHSIC ** 2 / varHSIC
	bet = varHSIC * n / mHSIC

	if mode == "pvalue":
		p_value = 1 - gamma.cdf(testStat, al, scale=bet)
		return p_value

	thresh = gamma.ppf(1 - alph, al, scale=bet)[0][0]

	if mode == "testStatMinusThres":
		return (testStat - thresh)

	return (testStat < thresh)


import networkx as nx
import numpy as np
import pandas as pd
from itertools import combinations, permutations
from copy import deepcopy

from pytrad.graph.GeneralGraph import GeneralGraph
from pytrad.graph.GraphNode import GraphNode
from pytrad.graph.Edge import Edge
from pytrad.graph.Endpoint import Endpoint
from pytrad.utils.GraphUtils import GraphUtils

from pytrad.utils.cit import fisherz

def trans_nodeset2str(Z):
    return tuple(str(z) for z in Z)


def is_fully_directed(edge):
    if edge:
        if edge.get_endpoint1() == Endpoint.TAIL and edge.get_endpoint2() == Endpoint.ARROW:
            return True
    return False


def is_endpoint(edge, z, end):
    if edge.get_node1() == z:
        if edge.get_endpoint1() == end:
            return True
        else:
            return False
    elif edge.get_node2() == z:
        if edge.get_endpoint2() == end:
            return True
        else:
            return False
    else:
        raise ValueError("z not in edge")


def mod_endpoint(edge, z, end):
    if edge.get_node1() == z:
        edge.set_endpoint1(end)
    elif edge.get_node2() == z:
        edge.set_endpoint2(end)
    else:
        raise ValueError("z not in edge")


# Zhang, Jiji. "A characterization of markov equivalence classes for directed acyclic graphs with latent variables."
def rule1(G):
    for x, y, z in combinations(G.get_nodes(), 3):
        if G.get_edge(x, z):
            continue
        edgexy = G.get_edge(x, y)
        edgeyz = G.get_edge(y, z)
        if not edgexy or not edgeyz:
            continue

        if is_endpoint(edgexy, y, Endpoint.ARROW) and is_endpoint(edgeyz, y, Endpoint.CIRCLE):
            G.remove_edge(edgeyz)
            mod_endpoint(edgeyz, y, Endpoint.TAIL)
            mod_endpoint(edgeyz, z, Endpoint.ARROW)
            print(f"Orient {y} --> {z}")
            G.add_edge(edgeyz)


def rule2(G):
    for x, y, z in combinations(G.get_nodes(), 3):
        if not G.get_edge(x, y) or not G.get_edge(x, z) or not G.get_edge(y, z):
            continue
        edgexy = G.get_edge(x, y)
        edgexz = G.get_edge(x, z)
        edgeyz = G.get_edge(y, z)

        if (is_fully_directed(edgexy) and is_endpoint(edgeyz, z, Endpoint.ARROW)) or \
                (is_fully_directed(edgeyz) and is_endpoint(edgexy, y, Endpoint.ARROW)):
            if is_endpoint(edgexz, z, Endpoint.CIRCLE):
                G.remove_edge(edgexz)
                mod_endpoint(edgexz, z, Endpoint.ARROW)
                print(f"Orient {x} *-> {z}")
                G.add_edge(edgexz)


def rule3(G):
    for x, y, z, w in combinations(G.get_nodes(), 4):
        edgexy = G.get_edge(x, y)
        edgexz = G.get_edge(x, z)
        edgexw = G.get_edge(x, w)
        edgeyz = G.get_edge(y, z)
        edgeyw = G.get_edge(y, w)
        edgezw = G.get_edge(z, w)

        if edgexz:
            continue
        cont = False

        for edge in [edgexy, edgexw, edgeyz, edgeyw, edgezw]:
            if not edge:
                cont = True
                break

        if cont:
            continue

        chain1 = is_endpoint(edgexy, y, Endpoint.ARROW) and is_endpoint(edgeyz, y, Endpoint.ARROW)
        chain2 = is_endpoint(edgexw, w, Endpoint.CIRCLE) and is_endpoint(edgezw, w, Endpoint.CIRCLE)
        if chain1 and chain2 and is_endpoint(edgeyw, y, Endpoint.CIRCLE):
            G.remove_edge(edgeyw)
            mod_endpoint(edgeyw, y, Endpoint.ARROW)
            print(f"Orient {w} *-> {y}")
            G.add_edge(edgeyw)


def rule4(G, sepset):
    nxG = nx.Graph()
    nodes = G.get_nodes()
    nodes_ids = {node: i for i, node in enumerate(nodes)}
    n = len(nodes)

    for i in range(n):
        nxG.add_node(i)
    for edge in G.get_graph_edges():
        nxG.add_edge(nodes_ids[edge.get_node1()], nodes_ids[edge.get_node2()])

    nodes = G.get_nodes()

    def has_directed_edge(prev, next):
        edge = G.get_edge(nodes[prev], nodes[next])
        if edge:
            if is_endpoint(edge, nodes[next], Endpoint.ARROW):
                return True
            else:
                return False
        else:
            return False

    def get_discriminating_paths(u, v, b):
        all_paths = nx.all_simple_paths(nxG, u, v)
        discpaths = []
        for path in all_paths:
            if b in path:
                b_pred = (path.index(v) - path.index(b)) == 1 and nxG.has_edge(b, v)
                all_colliders = True
                for node in path[1:-1]:
                    prev = path[path.index(node) - 1]
                    suc = path[path.index(node) + 1]
                    if (node != b) and not ((has_directed_edge(prev, node)) and (has_directed_edge(suc, node))):
                        all_colliders = False
                all_pred = True
                for node in path[1:-2]:
                    if not (has_directed_edge(node, v)):
                        all_pred = False
                nonadj = not nxG.has_edge(u, v)
                if (b_pred and all_colliders and all_pred and nonadj):
                    discpaths.append(path)
        return discpaths

    for x, y, z, w in combinations(nodes, 4):
        paths = get_discriminating_paths(nodes_ids[z], nodes_ids[w], nodes_ids[y])
        for path in paths:
            if nodes_ids[x] in path:
                edge_yz = G.get_edge(y, z)
                if not edge_yz:
                    continue
                if path.index(nodes_ids[x]) == len(path) - 3 and is_endpoint(edge_yz, y, Endpoint.CIRCLE):
                    if y in sepset[(z, w)]:
                        G.remove_edge(edge_yz)
                        mod_endpoint(edge_yz, y, Endpoint.TAIL)
                        mod_endpoint(edge_yz, z, Endpoint.ARROW)
                        G.add_edge(edge_yz)
                        print(f"Orient {y} --> {z}")
                    else:
                        edge_xy = G.get_edge(x, y)
                        if edge_xy:
                            G.remove_edge(edge_xy)
                            mod_endpoint(edge_xy, x, Endpoint.ARROW)
                            mod_endpoint(edge_xy, y, Endpoint.ARROW)
                            G.add_edge(edge_xy)
                            print(f"Orient {x} <-> {y}")
                        G.remove_edge(edge_yz)
                        mod_endpoint(edge_yz, y, Endpoint.ARROW)
                        mod_endpoint(edge_yz, z, Endpoint.ARROW)
                        G.add_edge(edge_yz)
                        print(f"Orient {y} <-> {z}")


def fci(data, indep_test=fisherz, alpha=0.05, verbose=False):
    '''
    Causal Discovery with Fast Causal Inference

    Parameters
    ----------
    data : Input data matrix
    indep_test : Independence test method function
    alpha : Significance level of individual partial correlation tests
    verbose: 0 - no output, 1 - detailed output

    Returns
    -------
    G : Causal graph
    '''

    assert type(data) == np.ndarray
    assert 0 < alpha < 1
    # assert test_name in ["Fisher_Z", "Chi_sq", "G_sq", "KCI"]

    no_of_var = data.shape[1]

    data = (data - np.mean(data, axis=0)) / np.std(data, axis=0)

    # corr_mat = None
    # if test_name == "Fisher_Z":
    #     corr_mat = np.corrcoef(data.T)
    #
    # kci_uind = None
    # kci_cind = None
    # if test_name == "KCI":
    #     kci_uind = KCI_UInd(sample_size=data.shape[0], kernelX=GaussianKernel(), kernelY=GaussianKernel(), approx=False)
    #     kci_cind = KCI_CInd(sample_size=data.shape[0], kernelX=GaussianKernel(), kernelY=GaussianKernel(),
    #                         kernelZ=GaussianKernel(), approx=False)

    nodes = []
    for i in range(no_of_var):
        node = GraphNode(f"X{i + 1}")
        node.add_attribute("id", i)
        nodes.append(node)
    G = GeneralGraph(nodes)

    sepset = {(nodex, nodey): set() for nodex, nodey in permutations(nodes, 2)}

    for x, y in combinations(nodes, 2):
        edge = Edge(x, y, Endpoint.CIRCLE, Endpoint.CIRCLE)
        edge.set_endpoint1(Endpoint.CIRCLE)
        edge.set_endpoint2(Endpoint.CIRCLE)
        G.add_edge(edge)

    depth = 0
    while depth <= no_of_var - 2:
        edge_removal = []
        for x, y in permutations(nodes, 2):
            adj_x = G.get_adjacent_nodes(x)
            if y in adj_x:
                adj_x.remove(y)
            else:
                continue

            if len(adj_x) < depth:
                continue
            for Z in combinations(adj_x, depth):
                S = tuple([i["id"] for i in Z])
                pval = indep_test(data, x["id"], y["id"], S)
                # if test_name == "Fisher_Z":
                #     pval = fisherZ(corr_mat, x["id"], y["id"], S, data.shape[0])
                # elif test_name == "Chi_sq":
                #     pval = chisq(data, x["id"], y["id"], S, G_sq=False)
                # elif test_name == "G_sq":
                #     pval = chisq(data, x["id"], y["id"], S, G_sq=True)
                # elif test_name == "KCI":
                #     if len(S) == 0:
                #         pval = kci_uind.compute_pvalue(data[:, x["id"]], data[:, y["id"]])
                #     else:
                #         pval = kci_cind.compute_pvalue(data[:, x["id"]], data[:, y["id"]], data[:, S])
                # else:
                #     raise NotImplementedError()
                if pval > alpha:
                    if verbose:
                        print(f"phase1 remove {x} --- {y} by sepset {trans_nodeset2str(Z)} pval:{pval}")
                    edge_removal.append(G.get_edge(x, y))
                    sepset[(x, y)] |= set(Z)
                    sepset[(y, x)] |= set(Z)

        for edge in edge_removal:
            if edge:
                G.remove_edge(edge)

        depth += 1

    for x, y in combinations(nodes, 2):
        for z in nodes:
            if x == z or y == z:
                continue
            if (not G.get_edge(x, y)) and G.get_edge(x, z) and G.get_edge(y, z) and z not in sepset[(x, y)]:
                edge_xz = G.get_edge(x, z)
                G.remove_edge(edge_xz)
                mod_endpoint(edge_xz, z, Endpoint.ARROW)
                G.add_edge(edge_xz)

                edge_yz = G.get_edge(y, z)
                G.remove_edge(edge_yz)
                mod_endpoint(edge_yz, z, Endpoint.ARROW)
                G.add_edge(edge_yz)

    nxG = nx.Graph()

    for node in G.get_nodes():
        nxG.add_node(node["id"])

    for edge in G.get_graph_edges():
        nxG.add_edge(edge.get_node1()["id"], edge.get_node2()["id"])

    def is_possible_d_sep(nxG, x, y):
        all_paths = nx.all_simple_paths(nxG, x, y)
        for path in all_paths:
            path_sep = True
            for i in range(1, len(path) - 1):
                is_collider = is_endpoint(G.get_edge(nodes[path[i - 1]], nodes[path[i]]), nodes[path[i]],
                                          Endpoint.ARROW) and \
                              is_endpoint(G.get_edge(nodes[path[i + 1]], nodes[path[i]]), nodes[path[i]],
                                          Endpoint.ARROW)
                is_triangle = G.get_edge(nodes[path[i - 1]], nodes[path[i + 1]]) is not None
                if not (is_collider or is_triangle):
                    path_sep = False
            if path_sep:
                return True
        return False

    pdsepset = {node: set() for node in nodes}

    for nodex, nodey in combinations(nodes, 2):
        if is_possible_d_sep(nxG, nodex["id"], nodey["id"]):
            pdsepset[nodex] |= {nodey}
            pdsepset[nodey] |= {nodex}

    # Phase 2
    G = GeneralGraph(nodes)

    for x, y in combinations(nodes, 2):
        edge = Edge(x, y, Endpoint.CIRCLE, Endpoint.CIRCLE)
        edge.set_endpoint1(Endpoint.CIRCLE)
        edge.set_endpoint2(Endpoint.CIRCLE)
        G.add_edge(edge)

    depth = 0
    while depth <= no_of_var - 2:
        edge_removal = []
        for x, y in permutations(nodes, 2):
            if not G.get_edge(x, y):
                continue

            pdseps = deepcopy(pdsepset[x])
            if y in pdseps:
                pdseps.remove(y)
            pdseps |= sepset[(x, y)]
            if len(pdseps) < depth:
                continue
            for Z in combinations(pdseps, depth):
                S = tuple([i["id"] for i in Z])
                # pval = 0.0
                pval = indep_test(data, x["id"], y["id"], S)
                # if test_name == "Fisher_Z":
                #     pval = fisherZ(corr_mat, x["id"], y["id"], S, data.shape[0])
                # elif test_name == "Chi_sq":
                #     pval = chisq(data, x["id"], y["id"], S, G_sq=False)
                # elif test_name == "G_sq":
                #     pval = chisq(data, x["id"], y["id"], S, G_sq=True)
                # elif test_name == "KCI":
                #     if len(S) == 0:
                #         pval = kci_uind.compute_pvalue(data[:, x["id"]], data[:, y["id"]])
                #     else:
                #         pval = kci_cind.compute_pvalue(data[:, x["id"]], data[:, y["id"]], data[:, S])
                # else:
                #     raise NotImplementedError()
                if pval >= alpha:
                    if verbose:
                        print(f"phase2 remove {x} --- {y} by sepset {trans_nodeset2str(Z)} pval:{pval}")
                    edge_removal.append(G.get_edge(x, y))
                    sepset[(x, y)] |= set(Z)
                    sepset[(y, x)] |= set(Z)

        for edge in edge_removal:
            if edge is not None:
                G.remove_edge(edge)

        depth += 1

    for x, y in combinations(nodes, 2):
        for z in nodes:
            if x == z or y == z:
                continue
            if (not G.get_edge(x, y)) and G.get_edge(x, z) and G.get_edge(y, z) and (z not in sepset[(x, y)]):
                print(f"orient_V: {x},{z},{y}")
                edge_xz = G.get_edge(x, z)
                G.remove_edge(edge_xz)
                mod_endpoint(edge_xz, z, Endpoint.ARROW)
                G.add_edge(edge_xz)

                edge_yz = G.get_edge(y, z)
                G.remove_edge(edge_yz)
                mod_endpoint(edge_yz, z, Endpoint.ARROW)
                G.add_edge(edge_yz)

    rule1(G)
    rule2(G)
    rule3(G)
    rule4(G, sepset)
    return G



from pytrad.utils.KCI.Kernel import Kernel


class LinearKernel(Kernel):
    def __init__(self, is_sparse=False):
        Kernel.__init__(self)
        self.is_sparse = is_sparse

    def __str__(self):
        s = self.__class__.__name__ + "=["
        s += "" + Kernel.__str__(self)
        s += "]"
        return s

    def kernel(self, X, Y=None):
        """
        Computes the linear kerpy k(x,y)=x^T y for the given data
        X - samples on right hand side
        Y - samples on left hand side, can be None in which case its replaced by X
        """

        if Y is None:
            Y = X
        if self.is_sparse:
            return X.dot(Y.T).todense()
        else:
            return X.dot(Y.T)



import setuptools

with open('README.md', 'r') as fh:
    README = fh.read()

VERSION = 0.001

setuptools.setup(
    name='Pytrad',
    version=VERSION,
    author='CLeaR',
    description='Pytrad Python Package',
    long_description=README,
    long_description_content_type='text/markdown',
    install_requires=[
        'numpy',
        'scipy',
        'scikit-learn',
        'graphviz',
        'statsmodels',
        'pandas',
        'matplotlib',
        'pygraphviz',
        'networkx'
    ],
    url='https://github.com/cmu-phil/pytrad',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)


import itertools
from copy import deepcopy

import numpy.matlib

from pytrad.graph.Edge import Edge
from pytrad.score.LocalScoreFunction import *


def feval(parameters):
    if parameters[0] == 'local_score_CV_general':
        return local_score_cv_general(parameters[1], parameters[2], parameters[3], parameters[4])
    elif parameters[0] == 'local_score_marginal_general':
        return local_score_marginal_general(parameters[1], parameters[2], parameters[3], parameters[4])
    elif parameters[0] == 'local_score_CV_multi':
        return local_score_cv_multi(parameters[1], parameters[2], parameters[3], parameters[4])
    elif parameters[0] == 'local_score_marginal_multi':
        return local_score_marginal_multi(parameters[1], parameters[2], parameters[3], parameters[4])
    elif parameters[0] == 'local_score_BIC':
        return local_score_bic(parameters[1], parameters[2], parameters[3], parameters[4])
    elif parameters[0] == 'local_score_BDeu':
        return local_score_bic(parameters[1], parameters[2], parameters[3], parameters[4])
    else:
        raise Exception('Undefined function')


def kernel(x, xKern, theta):
    # KERNEL Compute the rbf kernel
    n2 = dist2(x, xKern)
    if (theta[0] == 0):
        theta[0] = 2 / np.median(n2[np.where(np.tril(n2) > 0)])
        theta_new = theta[0]
    wi2 = theta[0] / 2
    kx = theta[1] * np.exp(-n2 * wi2)
    bw_new = 1 / theta[0]
    return kx, bw_new


def Combinatorial(T0):
    # sub = Combinatorial (T0); % find all the sbusets of T0
    sub = []
    count = 0
    if (len(T0) == 0):
        sub.append(())  # a 1x0 empty matrix
    else:
        if (len(T0) == 1):
            sub.append(())
            sub.append(T0)  # when T0 is a scale, it is a special case!!
        else:
            for n in range(len(T0) + 1):
                for S in list(itertools.combinations(T0, n)):
                    sub.append(S)
    return sub


def score_g(Data, G, score_func, parameters):  # calculate the score for the current G
    # here G is a DAG
    score = 0
    for i, node in enumerate(G.get_nodes()):
        PA = G.get_parents(node)
        delta_score = feval([score_func, Data, i, PA, parameters])
        score = score + delta_score
    return score


def insert_validity_test1(G, i, j, T):
    # V=Insert_validity_test1(G, X, Y, T,1); % do validity test for the operator Insert; V=1 means valid, V=0 mean invalid;
    # here G is CPDAG
    V = 0

    # condition 1
    Tj = np.intersect1d(np.where(G.graph[:, j] == -1)[0], np.where(G.graph[j, :] == -1)[0])  # neighbors of Xj
    Ti = np.union1d(np.where(G.graph[:, i] != 0)[0], np.where(G.graph[i, :] != 0)[0])  # adjacent to Xi;
    NA = np.intersect1d(Tj, Ti)  # find the neighbours of Xj and are adjacent to Xi
    V = check_clique(G, list(np.union1d(NA, T).astype(int)))  # check whether it is a clique
    return V


def check_clique(G, subnode):  # check whether node subnode is a clique in G
    # here G is a CPDAG
    # the definition of clique here: a clique is defined in an undirected graph
    # when you ignore the directionality of any directed edges
    Gs = deepcopy(G.graph[np.ix_(subnode, subnode)])  # extract the subgraph
    ns = len(subnode)

    if (ns == 0):
        s = 1
    else:
        row, col = np.where(Gs == 1)
        Gs[row, col] = -1
        Gs[col, row] = -1
        if (np.all((np.eye(ns) - np.ones((ns, ns))) == Gs)):  # check whether it is a clique
            s = 1
        else:
            s = 0
    return s


def insert_validity_test2(G, i, j, T):
    # V=Insert_validity_test(G, X, Y, T,1); % do validity test for the operator Insert; V=1 means valid, V=0 mean invalid;
    # here G is CPDAG
    V = 0
    Tj = np.intersect1d(np.where(G.graph[:, j] == -1)[0], np.where(G.graph[j, :] == -1)[0])  # neighbors of Xj
    Ti = np.union1d(np.where(G.graph[i, :] != 0)[0], np.where(G.graph[:, i] != 0)[0])  # adjacent to Xi;
    NA = np.intersect1d(Tj, Ti)  # find the neighbours of Xj and are adjacent to Xi

    # condition 2: every semi-directed path from Xj to Xi contains a node in union(NA,T)
    # Note: EVERY!!
    s2 = insert_vc2_new(G, j, i, np.union1d(NA, T))
    if (s2):
        V = 1

    return V


def insert_vc2_new(G, j, i, NAT):  # validity test for condition 2 of Insert operator
    # here G is CPDAG
    # Use Depth-first-Search
    start = j
    target = i
    # stack(1)=start; % initialize the stack
    stack = [{'value': start, 'pa': {}}]
    sign = 1  # If every semi-pathway contains a node in NAT, than sign=1;
    count = 1

    while (len(stack)):
        top = stack[0]
        stack = stack[1:]  # pop
        if (top['value'] == target):  # if find the target, search that pathway to see whether NAT is in that pathway
            curr = top
            ss = 0
            while True:
                if (len(curr['pa'])):
                    if (curr['pa']['value'] in NAT):  # contains a node in NAT
                        ss = 1
                        break
                else:
                    break
                curr = curr['pa']
            if (not ss):  # do not include NAT
                sign = 0
                break
        else:
            child = np.concatenate((np.where(G.graph[:, top['value']] == 1)[0],
                                    np.intersect1d(np.where(G.graph[top['value'], :] == -1)[0],
                                                   np.where(G.graph[:, top['value']] == -1)[0])))
            sign_child = np.ones(len(child))
            # check each child, whether it has appeared before in the same pathway
            for k in range(len(child)):
                curr = top
                while True:
                    if (len(curr['pa'])):
                        if (curr['pa']['value'] == child[k]):
                            sign_child[k] = 0  # has appeared in that path before
                            break
                    else:
                        break
                    curr = curr['pa']

            for k in range(len(sign_child)):
                if (sign_child[k]):
                    stack.insert(0, {'value': child[k], 'pa': top})  # push
    return sign


def find_subset_include(s0, sub):
    # S = find_subset_include(sub(k),sub); %  find those subsets that include sub(k)
    if (len(s0) == 0 or len(sub) == 0):
        Idx = np.ones(len(sub))
    else:
        Idx = np.zeros(len(sub))
        for i in range(len(sub)):
            tmp = set(s0).intersection(set(sub[i]))
            if (len(tmp)):
                if (tmp == set(s0)):
                    Idx[i] = 1
    return Idx


def insert_changed_score(Data, G, i, j, T, record_local_score, score_func, parameters):
    # calculate the changed score after the insert operator: i->j
    Tj = np.intersect1d(np.where(G.graph[:, j] == -1)[0], np.where(G.graph[j, :] == -1)[0])  # neighbors of Xj
    Ti = np.union1d(np.where(G.graph[i, :] != 0)[0], np.where(G.graph[:, i] != 0)[0])  # adjacent to Xi;
    NA = np.intersect1d(Tj, Ti)  # find the neighbours of Xj and are adjacent to Xi
    Paj = np.where(G.graph[j, :] == 1)[0]  # find the parents of Xj
    # the function local_score() calculates the local score
    tmp1 = np.union1d(NA, T).astype(int)
    tmp2 = np.union1d(tmp1, Paj)
    tmp3 = np.union1d(tmp2, [i]).astype(int)

    # before you calculate the local score, firstly you search in the
    # "record_local_score", to see whether you have calculated it before
    r = len(record_local_score[j])
    s1 = 0
    s2 = 0

    for r0 in range(r):
        if (not np.setxor1d(record_local_score[j][r0][0:-1], tmp3).size):
            score1 = record_local_score[j][r0][-1]
            s1 = 1

        if (not np.setxor1d(record_local_score[j][r0][0:-1],
                            tmp2).size):  # notice the differnece between 0*0 empty matrix and 1*0 empty matrix
            score2 = record_local_score[j][r0][-1]
            s2 = 1
        else:
            if ((not np.setxor1d(record_local_score[j][r0][0:-1], [-1]).size) and (not tmp2.size)):
                score2 = record_local_score[j][r0][-1]
                s2 = 1

        if (s1 and s2):
            break

    if (not s1):
        score1 = feval([score_func, Data, j, tmp3, parameters])
        temp = list(tmp3)
        temp.append(score1)
        record_local_score[j].append(temp)

    if (not s2):
        score2 = feval([score_func, Data, j, tmp2, parameters])
        # r = len(record_local_score[j])
        if (len(tmp2) != 0):
            temp = list(tmp2)
            temp.append(score2)
            record_local_score[j].append(temp)
        else:
            record_local_score[j].append([-1, score2])

    chscore = score1 - score2
    desc = [i, j, T]
    return chscore, desc, record_local_score


def insert(G, i, j, T):
    # Insert operator
    # insert the directed edge Xi->Xj
    nodes = G.get_nodes()
    G.add_edge(Edge(nodes[i], nodes[j], -1, 1))

    for k in range(len(T)):  # directing the previous undirected edge between T and Xj as T->Xj
        if G.get_edge(nodes[T[k]], nodes[j]) is not None:
            G.remove_edge(G.get_edge(nodes[T[k]], nodes[j]))
        G.add_edge(Edge(nodes[T[k]], nodes[j], -1, 1))

    return G


def delete_validity_test(G, i, j, H):
    # V=Delete_validity_test(G, X, Y, H); % do validity test for the operator Delete; V=1 means valid, V=0 mean invalid;
    # here G is CPDAG
    V = 0

    # condition 1
    Hj = np.intersect1d(np.where(G.graph[:, j] == -1)[0], np.where(G.graph[j, :] == -1)[0])  # neighbors of Xj
    Hi = np.union1d(np.where(G.graph[i, :] != 0)[0], np.where(G.graph[:, i] != 0)[0])  # adjacent to Xi;
    NA = np.intersect1d(Hj, Hi)  # find the neighbours of Xj and are adjacent to Xi
    s1 = check_clique(G, list(set(NA) - set(H)))  # check whether it is a clique

    if (s1):
        V = 1

    return V


def delete_changed_score(Data, G, i, j, H, record_local_score, score_func, parameters):
    # calculate the changed score after the Delete operator
    Hj = np.intersect1d(np.where(G.graph[:, j] == -1)[0], np.where(G.graph[j, :] == -1)[0])  # neighbors of Xj
    Hi = np.union1d(np.where(G.graph[i, :] != 0)[0], np.where(G.graph[:, i] != 0)[0])  # adjacent to Xi;
    NA = np.intersect1d(Hj, Hi)  # find the neighbours of Xj and are adjacent to Xi
    Paj = np.union1d(np.where(G.graph[j, :] == 1)[0], [i])  # find the parents of Xj
    # the function local_score() calculates the local score
    tmp1 = set(NA) - set(H)
    tmp2 = set.union(tmp1, set(Paj))
    tmp3 = tmp2 - {i}

    # before you calculate the local score, firstly you search in the
    # "record_local_score", to see whether you have calculated it before
    r = len(record_local_score[j])
    s1 = 0
    s2 = 0

    for r0 in range(r):
        if (set(record_local_score[j][r0][0:-1]) == tmp3):
            score1 = record_local_score[j][r0][-1]
            s1 = 1

        if (set(record_local_score[j][r0][
                0:-1]) == tmp2):  # notice the differnece between 0*0 empty matrix and 1*0 empty matrix
            score2 = record_local_score[j][r0][-1]
            s2 = 1
        else:
            if ((set(record_local_score[j][r0][0:-1]) == {-1}) and len(tmp2) == 0):
                score2 = record_local_score[j][r0][-1]
                s2 = 1

        if (s1 and s2):
            break

    if (not s1):
        score1 = feval([score_func, Data, j, list(tmp3), parameters])
        temp = list(tmp3)
        temp.append(score1)
        record_local_score[j].append(temp)

    if (not s2):
        score2 = feval([score_func, Data, j, list(tmp2), parameters])
        r = len(record_local_score[j])
        if (len(tmp2) != 0):
            temp = list(tmp2)
            temp.append(score2)
            record_local_score[j].append(temp)
        else:
            record_local_score[j].append([-1, score2])

    chscore = score1 - score2
    desc = [i, j, H]
    return chscore, desc, record_local_score


def delete(G, i, j, H):
    # Delete operator
    nodes = G.get_nodes()
    if G.get_edge(nodes[i], nodes[j]) is not None:
        # delete the edge between Xi and Xj
        G.remove_edge(G.get_edge(nodes[i], nodes[j]))
    for k in range(len(H)):  # directing the previous undirected edge
        if G.get_edge(nodes[j], nodes[H[k]]) is not None:
            G.remove_edge(G.get_edge(nodes[j], nodes[H[k]]))
        if G.get_edge(nodes[i], nodes[H[k]]) is not None:
            G.remove_edge(G.get_edge(nodes[i], nodes[H[k]]))
        G.add_edge(Edge(nodes[j], nodes[H[k]], -1, 1))
        G.add_edge(Edge(nodes[i], nodes[H[k]], -1, 1))
    return G


def dist2(x, c):
    # DIST2	Calculates squared distance between two sets of points.
    #
    # Description
    # D = DIST2(X, C) takes two matrices of vectors and calculates the
    # squared Euclidean distance between them.  Both matrices must be of
    # the same column dimension.  If X has M rows and N columns, and C has
    # L rows and N columns, then the result has M rows and L columns.  The
    # I, Jth entry is the  squared distance from the Ith row of X to the
    # Jth row of C.
    #
    # See also
    # GMMACTIV, KMEANS, RBFFWD
    #

    ndata, dimx = x.shape
    ncentres, dimc = c.shape
    if (dimx != dimc):
        raise Exception('Data dimension does not match dimension of centres')

    n2 = (np.matlib.ones((ncentres, 1)) * np.sum(np.multiply(x, x).T, axis=0)).T + \
         np.matlib.ones((ndata, 1)) * np.sum(np.multiply(c, c).T, axis=0) - \
         2 * (x * c.T)

    # Rounding errors occasionally cause negative entries in n2
    n2[np.where(n2 < 0)] = 0
    return n2


def pdinv(A):
    # PDINV Computes the inverse of a positive definite matrix
    numData = A.shape[0]
    try:
        U = np.linalg.cholesky(A).T
        invU = np.eye(numData).dot(np.linalg.inv(U))
        Ainv = invU.dot(invU.T)
    except numpy.linalg.LinAlgError as e:
        warnings.warn('Matrix is not positive definite in pdinv, inverting using svd')
        u, s, vh = np.linalg.svd(A, full_matrices=True)
        Ainv = vh.T.dot(np.diag(1 / s)).dot(u.T)
    except Exception as e:
        raise e
    return np.matlib.asmatrix(Ainv)


from pytrad.graph.Graph import Graph

class AdjacencyConfusion:
    __adjFn = 0
    __adjTp = 0
    __adjFp = 0
    __adjTn = 0

    def __init__(self, truth:Graph, est:Graph):
        nodes = truth.get_nodes()
        nodes_name = [node.get_name() for node in nodes]
        for i in list(range(0, len(nodes))):
            for j in list(range(i+1,len(nodes))):
                estAdj = est.is_adjacent_to(est.get_node(nodes_name[i]), est.get_node(nodes_name[j]))
                truthAdj = truth.is_adjacent_to(truth.get_node(nodes_name[i]), truth.get_node(nodes_name[j]))

                if truthAdj and not estAdj:
                    self.__adjFn = self.__adjFn + 1
                elif estAdj and not truthAdj:
                    self.__adjFp = self.__adjTp + 1
                elif estAdj and truthAdj:
                    self.__adjTp = self.__adjTp + 1
                elif not estAdj and not truthAdj:
                    self.__adjFn = self.__adjFn + 1
        allEdges = truth.get_num_nodes() * (truth.get_num_nodes() - 1) / 2
        self.__adjTn = allEdges - self.__adjFn

    def get_adj_tp(self):
        return self.__adjTp

    def get_adj_fp(self):
        return self.__adjFp

    def get_adj_fn(self):
        return self.__adjFn

    def get_adj_tn(self):
        return self.__adjTn

    def get_adj_precision(self):
        return self.__adjTp/ (self.__adjTp + self.__adjFp)

    def get_adj_recall(self):
        return self.__adjTp/ (self.__adjTp + self.__adjFn)



import os, sys
BASE_DIR = os.path.join(os.path.dirname(__file__), '..')
sys.path.append(BASE_DIR)
import pandas as pd
import numpy as np
from statsmodels.tsa.stattools import grangercausalitytests
from sklearn.linear_model import LassoCV


class Granger(object):
    def __init__(self, maxlag=2):
        self.maxlag = maxlag

    def granger_test(self, data, test='ssr_ftest'):
        n, dim = data.shape
        p_value_matrix = np.zeros((dim, dim))
        for c in range(dim):
            for r in range(dim):
                test_result = grangercausalitytests(data[:, [r, c]], maxlag=self.maxlag, verbose=False)
                p_values = [round(test_result[i + 1][0][test][1], 4) for i in range(self.maxlag)]
                min_p_value = np.min(p_values)
                p_value_matrix[r, c] = min_p_value
        return p_value_matrix

    def granger_lasso(self, data, cv=5):

        n, dim = data.shape
        # stack data to form one-vs-all regression
        Y = data[self.maxlag:]
        X = np.hstack([data[self.maxlag-k:-k] for k in range(1, self.maxlag+1)])

        lasso_cv = LassoCV(cv=cv)
        coeff = np.zeros((dim, dim*self.maxlag))
        # Consider one variable after the other as target
        for i in range(dim):
            lasso_cv.fit(X, Y[:,i])
            coeff[i] = lasso_cv.coef_
        return coeff


'''
    File name: GIN.py
    Discription: Learning Hidden Causal Representation with GIN condition
    Author: ZhiyiHuang@DMIRLab, RuichuCai@DMIRLab
    From DMIRLab: https://dmir.gdut.edu.cn/
'''

import numpy as np
from itertools import combinations

from pytrad.graph.NodeType import NodeType
from pytrad.graph.GraphNode import GraphNode
from pytrad.graph.GeneralGraph import GeneralGraph
from pytrad.search.FCMBased.lingam.hsic import hsic_test_gamma
from collections import deque

def GIN(data):
    '''
    Learning causal structure of Latent Variables for Linear Non-Gaussian Latent Variable Model
    with Generalized Independent Noise Condition

    Parameters
    ----------
    data : numpy ndarray
           data set

    Returns
    -------
    G : general graph
        causal graph
    K : list
        causal order
    '''
    v_labels = list(range(data.shape[1]))
    v_set = set(v_labels)
    cov = np.cov(data.T)

    # Step 1: Finding Causal Clusters
    cluster_list = []
    min_cluster = {i: set() for i in v_set}
    min_dep_score = {i: 1e9 for i in v_set}
    for (x1, x2) in combinations(v_set, 2):
        x_set = {x1, x2}
        z_set = v_set - x_set
        dep_statistic = cal_dep_for_gin(data, cov, list(x_set), list(z_set))
        for i in x_set:
            if min_dep_score[i] > dep_statistic:
                min_dep_score[i] = dep_statistic
                min_cluster[i] = x_set
    for i in v_labels:
        cluster_list.append(list(min_cluster[i]))

    cluster_list = merge_overlaping_cluster(cluster_list)

    # Step 2: Learning the Causal Order of Latent Variables
    K = []
    while (len(cluster_list) != 0):
        root = find_root(data, cov, cluster_list, K)
        K.append(root)
        cluster_list.remove(root)

    latent_id = 1
    l_nodes = []
    G = GeneralGraph([])
    for cluster in K:
        l_node = GraphNode(f"L{latent_id}")
        l_node.set_node_type(NodeType.LATENT)
        l_nodes.append(l_node)
        G.add_node(l_node)
        for l in l_nodes:
            if l != l_node:
                G.add_directed_edge(l, l_node)
        for o in cluster:
            o_node = GraphNode(f"X{o+1}")
            G.add_node(o_node)
            G.add_directed_edge(l_node, o_node)
        latent_id += 1

    return G, K

def cal_dep_for_gin(data, cov, X, Z):
    '''
    Calculate the statistics of dependence via Generalized Independent Noise Condition

    Parameters
    ----------
    data : data set (numpy ndarray)
    cov : covariance matrix
    X : test set variables
    Z : condition set variables

    Returns
    -------
    sta : test statistic
    '''
    cov_m = cov[np.ix_(Z, X)]
    _, _, v = np.linalg.svd(cov_m)
    omega = v.T[:, -1]
    e_xz = np.dot(omega, data[:, X].T)

    sta = 0
    for i in Z:
        sta += hsic_test_gamma(e_xz, data[:, i])[0]
    sta /= len(Z)
    return sta


def find_root(data, cov, clusters, K):
    '''
    Find the causal order by statistics of dependence

    Parameters
    ----------
    data : data set (numpy ndarray)
    cov : covariance matrix
    clusters : clusters of observed variables
    K : causal order

    Returns
    -------
    root : latent root cause
    '''
    if len(clusters) == 1:
        return clusters[0]
    root = clusters[0]
    dep_statistic_score = 1e30
    for i in clusters:
        for j in clusters:
            if i == j:
                continue
            X = [i[0], j[0]]
            Z = []
            for k in range(1, len(i)):
                Z.append(i[k])

            if K:
                for k in K:
                    X.append(k[0])
                    Z.append(k[1])

            dep_statistic = cal_dep_for_gin(data, cov, X, Z)
            if dep_statistic < dep_statistic_score:
                dep_statistic_score = dep_statistic
                root = i

    return root

def _get_all_elements(S):
    result = set()
    for i in S:
        for j in i:
            result |= {j}
    return result

# merging cluster
def merge_overlaping_cluster(cluster_list):
    v_labels = _get_all_elements(cluster_list)
    cluster_dict = {i: -1 for i in v_labels}
    cluster_b = {i: [] for i in v_labels}
    cluster_len = 0
    for i in range(len(cluster_list)):
        for j in cluster_list[i]:
            cluster_b[j].append(i)

    visited = [False] * len(cluster_list)
    cont = True
    while cont:
        cont = False
        q = deque()
        for i, val in enumerate(visited):
            if not val:
                q.append(i)
                visited[i] = True
                break
        while q:
            top = q.popleft()
            for i in cluster_list[top]:
                cluster_dict[i] = cluster_len
                for j in cluster_b[i]:
                    if not visited[j]:
                        q.append(j)
                        visited[j] = True

        for i in visited:
            if not i:
                cont = True
                break
        cluster_len += 1

    cluster = [[] for _ in range(cluster_len)]
    for i in v_labels:
        cluster[cluster_dict[i]].append(i)

    return cluster




from pytrad.graph.GeneralGraph import GeneralGraph
from pytrad.graph.GraphNode import GraphNode
from pytrad.graph.Edge import Edge
from pytrad.graph.Endpoint import Endpoint


def txt2generalgraph(filename):
    G = GeneralGraph([])
    node_map = {}
    with open(filename, "r") as file:
        next_nodes_line = False
        for line in file.readlines():
            line = line.strip()
            words = line.split()
            if len(words) > 1 and words[1] == 'Nodes:':
                next_nodes_line = True
            elif len(line) > 0 and next_nodes_line:
                next_nodes_line = False
                nodes = line.split(';')
                # print(nodes)
                for node in nodes:
                    node_map[node] = GraphNode(node)
                    G.add_node(node_map[node])
            elif len(words) > 0 and words[0][-1] == '.':
                next_nodes_line = False
                node1 = words[1]
                node2 = words[3]
                end1 = words[2][0]
                end2 = words[2][-1]
                if end1 == '<':
                    end1 = '>'
                end1 = to_endpoint(end1)
                end2 = to_endpoint(end2)
                edge = Edge(node_map[node1], node_map[node2], Endpoint.CIRCLE, Endpoint.CIRCLE)
                mod_endpoint(edge, node_map[node1], end1)
                mod_endpoint(edge, node_map[node2], end2)
                G.add_edge(edge)
    return G


def to_endpoint(s):
    if s == 'o':
        return Endpoint.CIRCLE
    elif s == '>':
        return Endpoint.ARROW
    elif s == '-':
        return Endpoint.TAIL
    else:
        raise NotImplementedError

def mod_endpoint(edge, z, end):
    if edge.get_node1() == z:
        edge.set_endpoint1(end)
    elif edge.get_node2() == z:
        edge.set_endpoint2(end)
    else:
        raise ValueError("z not in edge")
